#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <string.h>

#include "lzw.h"

#define Q_N		8		/* number of fixed point fractional bits */
#define SCALE		(1 << Q_N)	/* fixed point scale factor */
#define RSTRATE		487		/* RSTRATE/SCALE approximates 1.9 */

struct lzw_encoder enc;
unsigned char buffer[2047];
unsigned char buffer2[2*65536];
const unsigned char *in;
unsigned char *out;
size_t count, insz, outsz;
size_t nread, nwrite;
int finalize_each = 0;
int simulate_nospace = 0;
int do_reset = 0;

void
finalize()
{
	int rc;

	out = buffer;
	outsz = sizeof(buffer);
	rc = lzw_end_of_data(&enc, &out, &outsz);
	assert(rc == LZW_OK);

	if (out != buffer) {
		count = out - buffer;
		rc = write(1, buffer, count);
		if (rc == -1) {
			fprintf(stderr, "final write error: %s\n", strerror(errno));
			exit(2);
		}
		nwrite += count;
	}
}

void
finalize_simulate_nospace()
{
	static int max = 0;
	int rc;

	out = buffer;
	outsz = 0;

	do {
		outsz++;
		if (outsz > max) {
			max = outsz;
			fprintf(stderr, "outsz max to %d\n", max);
		}
		assert(outsz <= sizeof(buffer)-(unsigned long)out);
		rc = lzw_end_of_data(&enc, &out, &outsz);
	} while (rc == LZW_NOSPACE);

	assert(rc == LZW_OK);

	if (out != buffer) {
		count = out - buffer;
		rc = write(1, buffer, count);
		if (rc == -1) {
			fprintf(stderr, "final write error: %s\n", strerror(errno));
			exit(2);
		}
		nwrite += count;
	}
}

int
main(int argc, char *argv[])
{
	int opt, rc;

	nread = nwrite = 0;
	opterr = 0;

	while ((opt = getopt(argc, argv, "irs")) != -1) {
		switch (opt) {
		case 'i':
			finalize_each = 1;
			break;
		case 'r':
			do_reset = 1;
			break;
		case 's':
			simulate_nospace = 1;	
			break;
		default:
			fprintf(stderr, "unknown option\n");
			exit(1);
		}
	}

	lzw_encoder_init(&enc);

	while ((count = read(0, buffer, sizeof(buffer))) > 0) {

		in = buffer;
		insz = count;
		nread += count;

		while (insz > 0) {

			out = buffer2;
			outsz = sizeof(buffer2);
			rc = lzw_encode(&enc, &in, &insz, &out, &outsz);

			if (rc == LZW_OK || rc == LZW_NOSPACE) {
				uint64_t res;

				res = count << Q_N;
				res += (out - buffer2) >> 1;
				if (out - buffer2 > 0)
					res /= (out - buffer2);
				else
					res = INT_MAX;

				if (do_reset && lzw_complete(&enc) && res < RSTRATE) {
					//fprintf(stderr, "reset!\n");
					rc = lzw_end_of_data(&enc, &out, &outsz);
					assert(rc == LZW_OK);
					rc = lzw_reset(&enc, &out, &outsz);
					assert(rc == LZW_OK);
				}

				count = out - buffer2;
				rc = write(1, buffer2, count);
				if (rc == -1) {
					fprintf(stderr, "write error: %s\n", strerror(errno));
					exit(2);
				}
				nwrite += count;
			} else {
				if (rc == LZW_NOMEM)
					fprintf(stderr, "out of memory\n");
				else
					fprintf(stderr, "error code %d\n", rc);
				exit(2);
			}
		}
		if (finalize_each) {
			if (simulate_nospace)
				finalize_simulate_nospace();
			else
				finalize();
		}
	}

	if (!finalize_each) {
		if (simulate_nospace)
			finalize_simulate_nospace();
		else
			finalize();
	}

	lzw_encoder_deinit(&enc);

	//fprintf(stderr, "nread %lu, nwrite %lu\n", nread, nwrite);

	return 0;
}

